# -*- encoding: utf-8 -*-
# Copyright (c) 2016 Servionica LTD
# Copyright (c) 2016 Intel Corp
#
# Authors: Alexander Chadin <a.chadin@servionica.ru>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime

from croniter import croniter
from dateutil import tz
from oslo_utils import timeutils

from watcher.common import context
from watcher.common import scheduling
from watcher.common import utils
from watcher import conf
from watcher.db.sqlalchemy import api as sq_api
from watcher.db.sqlalchemy import job_store
from watcher.decision_engine.audit import base
from watcher import objects


CONF = conf.CONF


class ContinuousAuditHandler(base.AuditHandler):
    def __init__(self):
        super(ContinuousAuditHandler, self).__init__()
        # scheduler for executing audits
        self._audit_scheduler = None
        # scheduler for a periodic task to launch audit
        self._period_scheduler = None
        self.context_show_deleted = context.RequestContext(is_admin=True,
                                                           show_deleted=True)

    @property
    def scheduler(self):
        if self._audit_scheduler is None:
            self._audit_scheduler = scheduling.BackgroundSchedulerService(
                jobstores={
                    'default': job_store.WatcherJobStore(
                        engine=sq_api.enginefacade.writer.get_engine()),
                }
            )
        return self._audit_scheduler

    @property
    def period_scheduler(self):
        if self._period_scheduler is None:
            self._period_scheduler = scheduling.BackgroundSchedulerService()
        return self._period_scheduler

    def _is_audit_inactive(self, audit):
        audit = objects.Audit.get_by_uuid(
            self.context_show_deleted, audit.uuid, eager=True)
        if (objects.audit.AuditStateTransitionManager().is_inactive(audit) or
                (audit.hostname != CONF.host) or
                (self.check_audit_expired(audit))):
            # if audit isn't in active states, audit's job must be removed to
            # prevent using of inactive audit in future.
            jobs = [job for job in self.scheduler.get_jobs()
                    if job.name == 'execute_audit' and
                    job.args[0].uuid == audit.uuid]
            if jobs:
                jobs[0].remove()
            return True

        return False

    def do_execute(self, audit, request_context):
        solution = super(ContinuousAuditHandler, self)\
            .do_execute(audit, request_context)

        if audit.audit_type == objects.audit.AuditType.CONTINUOUS.value:
            a_plan_filters = {'audit_uuid': audit.uuid,
                              'state': objects.action_plan.State.RECOMMENDED}
            action_plans = objects.ActionPlan.list(
                request_context, filters=a_plan_filters, eager=True)
            for plan in action_plans:
                plan.state = objects.action_plan.State.CANCELLED
                plan.save()
        return solution

    @staticmethod
    def _next_cron_time(audit):
        if utils.is_cron_like(audit.interval):
            return croniter(audit.interval, timeutils.utcnow()
                            ).get_next(datetime.datetime)

    @classmethod
    def execute_audit(cls, audit, request_context):
        self = cls()
        if not self._is_audit_inactive(audit):
            try:
                self.execute(audit, request_context)
            except Exception:
                raise
            finally:
                if utils.is_int_like(audit.interval):
                    audit.next_run_time = (
                        timeutils.utcnow() +
                        datetime.timedelta(seconds=int(audit.interval)))
                else:
                    audit.next_run_time = self._next_cron_time(audit)
                audit.save()

    def _add_job(self, trigger, audit, audit_context, **trigger_args):
        time_var = 'next_run_time' if trigger_args.get(
            'next_run_time') else 'run_date'
        # We should convert UTC time to local time without tzinfo
        trigger_args[time_var] = trigger_args[time_var].replace(
            tzinfo=tz.tzutc()).astimezone(tz.tzlocal()).replace(tzinfo=None)
        self.scheduler.add_job(self.execute_audit, trigger,
                               args=[audit, audit_context],
                               name='execute_audit',
                               **trigger_args)

    def check_audit_expired(self, audit):
        current = timeutils.utcnow()
        # Note: if audit still didn't get into the timeframe,
        #       skip it
        if audit.start_time and audit.start_time > current:
            return True
        if audit.end_time and audit.end_time < current:
            if audit.state != objects.audit.State.SUCCEEDED:
                audit.state = objects.audit.State.SUCCEEDED
                audit.save()
            return True

        return False

    def launch_audits_periodically(self):
        # if audit scheduler stop, restart it
        if not self.scheduler.running:
            self.scheduler.start()

        audit_context = context.RequestContext(is_admin=True)
        audit_filters = {
            'audit_type': objects.audit.AuditType.CONTINUOUS.value,
            'state__in': (objects.audit.State.PENDING,
                          objects.audit.State.ONGOING),
        }
        audit_filters['hostname'] = None
        unscheduled_audits = objects.Audit.list(
            audit_context, filters=audit_filters, eager=True)
        for audit in unscheduled_audits:
            # If continuous audit doesn't have a hostname yet,
            # Watcher will set current CONF.host value.
            # TODO(alexchadin): Add scheduling of new continuous audits.
            audit.hostname = CONF.host
            audit.save()
        scheduler_job_args = [
            (job.args[0].uuid, job) for job
            in self.scheduler.get_jobs()
            if job.name == 'execute_audit']
        scheduler_jobs = dict(scheduler_job_args)
        # if audit isn't in active states, audit's job should be removed
        jobs_to_remove = []
        for job in scheduler_jobs.values():
            if self._is_audit_inactive(job.args[0]):
                jobs_to_remove.append(job.args[0].uuid)
        for audit_uuid in jobs_to_remove:
            scheduler_jobs.pop(audit_uuid)
        audit_filters['hostname'] = CONF.host
        audits = objects.Audit.list(
            audit_context, filters=audit_filters, eager=True)
        for audit in audits:
            if self.check_audit_expired(audit):
                continue
            existing_job = scheduler_jobs.get(audit.uuid, None)
            # if audit is not presented in scheduled audits yet,
            # just add a new audit job.
            # if audit is already in the job queue, and interval has changed,
            # we need to remove the old job and add a new one.
            if (existing_job is None) or (
                existing_job and
                    audit.interval != existing_job.args[0].interval):
                if existing_job:
                    self.scheduler.remove_job(existing_job.id)
                # if interval is provided with seconds
                if utils.is_int_like(audit.interval):
                    # if audit has already been provided and we need
                    # to restore it after shutdown
                    if audit.next_run_time is not None:
                        old_run_time = audit.next_run_time
                        current = timeutils.utcnow()
                        if old_run_time < current:
                            delta = datetime.timedelta(
                                seconds=(int(audit.interval) - (
                                    current - old_run_time).seconds %
                                    int(audit.interval)))
                            audit.next_run_time = current + delta
                        next_run_time = audit.next_run_time
                    # if audit is new one
                    else:
                        next_run_time = timeutils.utcnow()
                    self._add_job('interval', audit, audit_context,
                                  seconds=int(audit.interval),
                                  next_run_time=next_run_time)

                else:
                    audit.next_run_time = self._next_cron_time(audit)
                    self._add_job('date', audit, audit_context,
                                  run_date=audit.next_run_time)
                audit.hostname = CONF.host
                audit.save()

    def start(self):
        self.period_scheduler.add_job(
            self.launch_audits_periodically,
            'interval',
            seconds=CONF.watcher_decision_engine.continuous_audit_interval,
            next_run_time=datetime.datetime.now())
        self.period_scheduler.start()
        # audit scheduler start
        self.scheduler.start()



# Copyright 2023 Cloudbase Solutions
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from watcher.common import exception
from watcher.common.metal_helper import base as m_helper_base
from watcher.common.metal_helper import constants as m_constants
from watcher.tests import base


# The base classes have abstract methods, we'll need to
# stub them.
class MockMetalNode(m_helper_base.BaseMetalNode):
    def get_power_state(self):
        raise NotImplementedError()

    def get_id(self):
        raise NotImplementedError()

    def power_on(self):
        raise NotImplementedError()

    def power_off(self):
        raise NotImplementedError()


class MockMetalHelper(m_helper_base.BaseMetalHelper):
    def list_compute_nodes(self):
        pass

    def get_node(self, node_id):
        pass


class TestBaseMetalNode(base.TestCase):
    def setUp(self):
        super().setUp()

        self._nova_node = mock.Mock()
        self._node = MockMetalNode(self._nova_node)

    def test_get_hypervisor_node(self):
        self.assertEqual(
            self._nova_node,
            self._node.get_hypervisor_node())

    def test_get_hypervisor_node_missing(self):
        node = MockMetalNode()
        self.assertRaises(
            exception.Invalid,
            node.get_hypervisor_node)

    def test_get_hypervisor_hostname(self):
        self.assertEqual(
            self._nova_node.hypervisor_hostname,
            self._node.get_hypervisor_hostname())

    @mock.patch.object(MockMetalNode, 'power_on')
    @mock.patch.object(MockMetalNode, 'power_off')
    def test_set_power_state(self,
                             mock_power_off, mock_power_on):
        self._node.set_power_state(m_constants.PowerState.ON)
        mock_power_on.assert_called_once_with()

        self._node.set_power_state(m_constants.PowerState.OFF)
        mock_power_off.assert_called_once_with()

        self.assertRaises(
            exception.UnsupportedActionType,
            self._node.set_power_state,
            m_constants.PowerState.UNKNOWN)


class TestBaseMetalHelper(base.TestCase):
    def setUp(self):
        super().setUp()

        self._osc = mock.Mock()
        self._helper = MockMetalHelper(self._osc)

    def test_nova_client_attr(self):
        self.assertEqual(self._osc.nova.return_value,
                         self._helper.nova_client)



# -*- encoding: utf-8 -*-
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from apscheduler.schedulers import background
import freezegun
from oslo_utils import timeutils

from watcher.api import scheduling
from watcher.notifications import service
from watcher import objects
from watcher.tests import base
from watcher.tests.db import base as db_base
from watcher.tests.db import utils


class TestSchedulingService(base.TestCase):

    @mock.patch.object(background.BackgroundScheduler, 'start')
    def test_start_scheduling_service(self, m_start):
        scheduler = scheduling.APISchedulingService()
        scheduler.start()
        m_start.assert_called_once_with(scheduler)
        jobs = scheduler.get_jobs()
        self.assertEqual(1, len(jobs))


class TestSchedulingServiceFunctions(db_base.DbTestCase):

    def setUp(self):
        super(TestSchedulingServiceFunctions, self).setUp()
        fake_service = utils.get_test_service(
            created_at=timeutils.utcnow())
        self.fake_service = objects.Service(**fake_service)

    @mock.patch.object(scheduling.APISchedulingService, 'get_service_status')
    @mock.patch.object(objects.Service, 'list')
    @mock.patch.object(service, 'send_service_update')
    def test_get_services_status_without_services_in_list(
            self, mock_service_update, mock_get_list, mock_service_status):
        scheduler = scheduling.APISchedulingService()
        mock_get_list.return_value = [self.fake_service]
        mock_service_status.return_value = 'ACTIVE'
        scheduler.get_services_status(mock.ANY)
        mock_service_status.assert_called_once_with(mock.ANY,
                                                    self.fake_service.id)

        mock_service_update.assert_not_called()

    @mock.patch.object(scheduling.APISchedulingService, 'get_service_status')
    @mock.patch.object(objects.Service, 'list')
    @mock.patch.object(service, 'send_service_update')
    def test_get_services_status_with_services_in_list_same_status(
            self, mock_service_update, mock_get_list, mock_service_status):
        scheduler = scheduling.APISchedulingService()
        mock_get_list.return_value = [self.fake_service]
        scheduler.services_status = {1: 'ACTIVE'}
        mock_service_status.return_value = 'ACTIVE'
        scheduler.get_services_status(mock.ANY)
        mock_service_status.assert_called_once_with(mock.ANY,
                                                    self.fake_service.id)

        mock_service_update.assert_not_called()

    @mock.patch.object(scheduling.APISchedulingService, 'get_service_status')
    @mock.patch.object(objects.Service, 'list')
    @mock.patch.object(service, 'send_service_update')
    def test_get_services_status_with_services_in_list_diff_status(
            self, mock_service_update, mock_get_list, mock_service_status):
        scheduler = scheduling.APISchedulingService()
        mock_get_list.return_value = [self.fake_service]
        scheduler.services_status = {1: 'FAILED'}
        mock_service_status.return_value = 'ACTIVE'
        scheduler.get_services_status(mock.ANY)
        mock_service_status.assert_called_once_with(mock.ANY,
                                                    self.fake_service.id)

        mock_service_update.assert_called_once_with(mock.ANY,
                                                    self.fake_service,
                                                    state='ACTIVE')

    @mock.patch.object(objects.Service, 'get')
    def test_get_service_status_failed_service(
            self, mock_get):
        scheduler = scheduling.APISchedulingService()
        mock_get.return_value = self.fake_service
        service_status = scheduler.get_service_status(mock.ANY,
                                                      self.fake_service.id)
        mock_get.assert_called_once_with(mock.ANY,
                                         self.fake_service.id)
        self.assertEqual('FAILED', service_status)

    @freezegun.freeze_time('2016-09-22T08:32:26.219414')
    @mock.patch.object(objects.Service, 'get')
    def test_get_service_status_failed_active(
            self, mock_get):
        scheduler = scheduling.APISchedulingService()
        mock_get.return_value = self.fake_service
        service_status = scheduler.get_service_status(mock.ANY,
                                                      self.fake_service.id)
        mock_get.assert_called_once_with(mock.ANY,
                                         self.fake_service.id)
        self.assertEqual('ACTIVE', service_status)



# Copyright 2023 Cloudbase Solutions
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

try:
    from maas.client import enum as maas_enum
except ImportError:
    maas_enum = None

from watcher.common.metal_helper import constants as m_constants
from watcher.common.metal_helper import maas
from watcher.tests import base


class TestMaasNode(base.TestCase):
    def setUp(self):
        super().setUp()

        self._wrapped_node = mock.Mock()
        self._nova_node = mock.Mock()
        self._maas_client = mock.Mock()

        self._node = maas.MaasNode(
            self._wrapped_node, self._nova_node, self._maas_client)

    def test_get_power_state(self):
        if not maas_enum:
            self.skipTest("python-libmaas not installed.")

        self._wrapped_node.query_power_state.side_effect = (
            maas_enum.PowerState.ON,
            maas_enum.PowerState.OFF,
            maas_enum.PowerState.ERROR,
            maas_enum.PowerState.UNKNOWN,
            'SomeOtherState')

        expected_states = (
            m_constants.PowerState.ON,
            m_constants.PowerState.OFF,
            m_constants.PowerState.ERROR,
            m_constants.PowerState.UNKNOWN,
            m_constants.PowerState.UNKNOWN)

        for expected_state in expected_states:
            actual_state = self._node.get_power_state()
            self.assertEqual(expected_state, actual_state)

    def test_get_id(self):
        self.assertEqual(
            self._wrapped_node.system_id,
            self._node.get_id())

    def test_power_on(self):
        self._node.power_on()
        self._wrapped_node.power_on.assert_called_once_with()

    def test_power_off(self):
        self._node.power_off()
        self._wrapped_node.power_off.assert_called_once_with()


class TestMaasHelper(base.TestCase):
    def setUp(self):
        super().setUp()

        self._mock_osc = mock.Mock()
        self._mock_nova_client = self._mock_osc.nova.return_value
        self._mock_maas_client = self._mock_osc.maas.return_value
        self._helper = maas.MaasHelper(osc=self._mock_osc)

    def test_list_compute_nodes(self):
        compute_fqdn = "compute-0"
        # some other MAAS node, not a Nova node
        ctrl_fqdn = "ctrl-1"

        mock_machines = [
            mock.Mock(fqdn=compute_fqdn,
                      system_id=mock.sentinel.compute_node_id),
            mock.Mock(fqdn=ctrl_fqdn,
                      system_id=mock.sentinel.ctrl_node_id),
        ]
        mock_hypervisors = [
            mock.Mock(hypervisor_hostname=compute_fqdn),
        ]

        self._mock_maas_client.machines.list.return_value = mock_machines
        self._mock_nova_client.hypervisors.list.return_value = mock_hypervisors

        out_nodes = self._helper.list_compute_nodes()
        self.assertEqual(1, len(out_nodes))

        out_node = out_nodes[0]
        self.assertIsInstance(out_node, maas.MaasNode)
        self.assertEqual(mock.sentinel.compute_node_id, out_node.get_id())
        self.assertEqual(compute_fqdn, out_node.get_hypervisor_hostname())

    def test_get_node(self):
        mock_machine = mock.Mock(fqdn='compute-0')
        self._mock_maas_client.machines.get.return_value = mock_machine

        mock_compute_nodes = [
            mock.Mock(hypervisor_hostname="compute-011"),
            mock.Mock(hypervisor_hostname="compute-0"),
            mock.Mock(hypervisor_hostname="compute-01"),
        ]
        self._mock_nova_client.hypervisors.search.return_value = (
            mock_compute_nodes)

        out_node = self._helper.get_node(mock.sentinel.id)

        self.assertEqual(mock_compute_nodes[1], out_node._nova_node)
        self.assertEqual(self._mock_maas_client, out_node._maas_client)
        self.assertEqual(mock_machine, out_node._maas_node)



# -*- encoding: utf-8 -*-
# Copyright (c) 2017 ZTE
#
# Authors: Li Canwei
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import time

from oslo_log import log

from watcher._i18n import _
from watcher.applier.actions import base
from watcher.common import exception
from watcher.common.metal_helper import constants as metal_constants
from watcher.common.metal_helper import factory as metal_helper_factory

LOG = log.getLogger(__name__)


class ChangeNodePowerState(base.BaseAction):
    """Compute node power on/off

    By using this action, you will be able to on/off the power of a
    compute node.

    The action schema is::

        schema = Schema({
         'resource_id': str,
         'state': str,
        })

    The `resource_id` references a baremetal node id (list of available
    ironic nodes is returned by this command: ``ironic node-list``).
    The `state` value should either be `on` or `off`.
    """

    STATE = 'state'

    @property
    def schema(self):
        return {
            'type': 'object',
            'properties': {
                'resource_id': {
                    'type': 'string',
                    "minlength": 1
                },
                'resource_name': {
                    'type': 'string',
                    "minlength": 1
                },
                'state': {
                    'type': 'string',
                    'enum': [metal_constants.PowerState.ON.value,
                             metal_constants.PowerState.OFF.value]
                }
            },
            'required': ['resource_id', 'state'],
            'additionalProperties': False,
        }

    @property
    def node_uuid(self):
        return self.resource_id

    @property
    def state(self):
        return self.input_parameters.get(self.STATE)

    def execute(self):
        target_state = self.state
        return self._node_manage_power(target_state)

    def revert(self):
        if self.state == metal_constants.PowerState.ON.value:
            target_state = metal_constants.PowerState.OFF.value
        elif self.state == metal_constants.PowerState.OFF.value:
            target_state = metal_constants.PowerState.ON.value
        return self._node_manage_power(target_state)

    def _node_manage_power(self, state, retry=60):
        if state is None:
            raise exception.IllegalArgumentException(
                message=_("The target state is not defined"))

        metal_helper = metal_helper_factory.get_helper(self.osc)
        node = metal_helper.get_node(self.node_uuid)
        current_state = node.get_power_state()

        if state == current_state.value:
            return True

        if state == metal_constants.PowerState.OFF.value:
            compute_node = node.get_hypervisor_node().to_dict()
            if (compute_node['running_vms'] == 0):
                node.set_power_state(state)
            else:
                LOG.warning(
                    "Compute node %s has %s running vms and will "
                    "NOT be shut off.",
                    compute_node["hypervisor_hostname"],
                    compute_node['running_vms'])
                return False
        else:
            node.set_power_state(state)

        node = metal_helper.get_node(self.node_uuid)
        while node.get_power_state() == current_state and retry:
            time.sleep(10)
            retry -= 1
            node = metal_helper.get_node(self.node_uuid)
        if retry > 0:
            return True
        else:
            return False

    def pre_condition(self):
        pass

    def post_condition(self):
        pass

    def get_description(self):
        """Description of the action"""
        return ("Compute node power on/off through Ironic or MaaS.")



# Copyright 2023 Cloudbase Solutions
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from watcher.common.metal_helper import constants as m_constants
from watcher.common.metal_helper import ironic
from watcher.tests import base


class TestIronicNode(base.TestCase):
    def setUp(self):
        super().setUp()

        self._wrapped_node = mock.Mock()
        self._nova_node = mock.Mock()
        self._ironic_client = mock.Mock()

        self._node = ironic.IronicNode(
            self._wrapped_node, self._nova_node, self._ironic_client)

    def test_get_power_state(self):
        states = (
            "power on",
            "power off",
            "rebooting",
            "soft power off",
            "soft reboot",
            'SomeOtherState')
        type(self._wrapped_node).power_state = mock.PropertyMock(
            side_effect=states)

        expected_states = (
            m_constants.PowerState.ON,
            m_constants.PowerState.OFF,
            m_constants.PowerState.ON,
            m_constants.PowerState.OFF,
            m_constants.PowerState.ON,
            m_constants.PowerState.UNKNOWN)

        for expected_state in expected_states:
            actual_state = self._node.get_power_state()
            self.assertEqual(expected_state, actual_state)

    def test_get_id(self):
        self.assertEqual(
            self._wrapped_node.uuid,
            self._node.get_id())

    def test_power_on(self):
        self._node.power_on()
        self._ironic_client.node.set_power_state.assert_called_once_with(
            self._wrapped_node.uuid, "on")

    def test_power_off(self):
        self._node.power_off()
        self._ironic_client.node.set_power_state.assert_called_once_with(
            self._wrapped_node.uuid, "off")


class TestIronicHelper(base.TestCase):
    def setUp(self):
        super().setUp()

        self._mock_osc = mock.Mock()
        self._mock_nova_client = self._mock_osc.nova.return_value
        self._mock_ironic_client = self._mock_osc.ironic.return_value
        self._helper = ironic.IronicHelper(osc=self._mock_osc)

    def test_list_compute_nodes(self):
        mock_machines = [
            mock.Mock(
                extra=dict(compute_node_id=mock.sentinel.compute_node_id)),
            mock.Mock(
                extra=dict(compute_node_id=mock.sentinel.compute_node_id2)),
            mock.Mock(
                extra=dict())
        ]
        mock_hypervisor = mock.Mock()

        self._mock_ironic_client.node.list.return_value = mock_machines
        self._mock_ironic_client.node.get.side_effect = mock_machines
        self._mock_nova_client.hypervisors.get.side_effect = (
            mock_hypervisor, None)

        out_nodes = self._helper.list_compute_nodes()
        self.assertEqual(1, len(out_nodes))

        out_node = out_nodes[0]
        self.assertIsInstance(out_node, ironic.IronicNode)
        self.assertEqual(mock_hypervisor, out_node._nova_node)
        self.assertEqual(mock_machines[0], out_node._ironic_node)
        self.assertEqual(self._mock_ironic_client, out_node._ironic_client)

    def test_get_node(self):
        mock_machine = mock.Mock(
            extra=dict(compute_node_id=mock.sentinel.compute_node_id))
        self._mock_ironic_client.node.get.return_value = mock_machine

        out_node = self._helper.get_node(mock.sentinel.id)

        self.assertEqual(self._mock_nova_client.hypervisors.get.return_value,
                         out_node._nova_node)
        self.assertEqual(self._mock_ironic_client, out_node._ironic_client)
        self.assertEqual(mock_machine, out_node._ironic_node)

    def test_get_node_not_a_hypervisor(self):
        mock_machine = mock.Mock(extra=dict(compute_node_id=None))
        self._mock_ironic_client.node.get.return_value = mock_machine

        out_node = self._helper.get_node(mock.sentinel.id)

        self._mock_nova_client.hypervisors.get.assert_not_called()
        self.assertIsNone(out_node._nova_node)
        self.assertEqual(self._mock_ironic_client, out_node._ironic_client)
        self.assertEqual(mock_machine, out_node._ironic_node)



# Copyright 2023 Cloudbase Solutions
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg

maas_client = cfg.OptGroup(name='maas_client',
                           title='Configuration Options for MaaS')

MAAS_CLIENT_OPTS = [
    cfg.StrOpt('url',
               help='MaaS URL, example: http://1.2.3.4:5240/MAAS'),
    cfg.StrOpt('api_key',
               help='MaaS API authentication key.'),
    cfg.IntOpt('timeout',
               default=60,
               help='MaaS client operation timeout in seconds.')]


def register_opts(conf):
    conf.register_group(maas_client)
    conf.register_opts(MAAS_CLIENT_OPTS, group=maas_client)


def list_opts():
    return [(maas_client, MAAS_CLIENT_OPTS)]



# -*- encoding: utf-8 -*-
# Copyright (c) 2015 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities and helper functions."""

import asyncio
import datetime
import inspect
import random
import re
import string

from croniter import croniter
import eventlet
from eventlet import tpool

from jsonschema import validators
from oslo_config import cfg
from oslo_log import log
from oslo_utils import strutils
from oslo_utils import uuidutils

from watcher.common import exception

CONF = cfg.CONF

LOG = log.getLogger(__name__)


class Struct(dict):
    """Specialized dict where you access an item like an attribute

    >>> struct = Struct()
    >>> struct['a'] = 1
    >>> struct.b = 2
    >>> assert struct.a == 1
    >>> assert struct['b'] == 2
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setattr__(self, name, value):
        try:
            self[name] = value
        except KeyError:
            raise AttributeError(name)


generate_uuid = uuidutils.generate_uuid
is_uuid_like = uuidutils.is_uuid_like
is_int_like = strutils.is_int_like


def is_cron_like(value):
    """Return True is submitted value is like cron syntax"""
    try:
        croniter(value, datetime.datetime.now())
    except Exception as e:
        raise exception.CronFormatIsInvalid(message=str(e))
    return True


def safe_rstrip(value, chars=None):
    """Removes trailing characters from a string if that does not make it empty

    :param value: A string value that will be stripped.
    :param chars: Characters to remove.
    :return: Stripped value.

    """
    if not isinstance(value, str):
        LOG.warning(
            "Failed to remove trailing character. Returning original object."
            "Supplied object is not a string: %s,", value)
        return value

    return value.rstrip(chars) or value


def is_hostname_safe(hostname):
    """Determine if the supplied hostname is RFC compliant.

    Check that the supplied hostname conforms to:
        * http://en.wikipedia.org/wiki/Hostname
        * http://tools.ietf.org/html/rfc952
        * http://tools.ietf.org/html/rfc1123

    :param hostname: The hostname to be validated.
    :returns: True if valid. False if not.

    """
    m = r'^[a-z0-9]([a-z0-9\-]{0,61}[a-z0-9])?$'
    return (isinstance(hostname, str) and
            (re.match(m, hostname) is not None))


def get_cls_import_path(cls):
    """Return the import path of a given class"""
    module = cls.__module__
    if module is None or module == str.__module__:
        return cls.__name__
    return module + '.' + cls.__name__


# Default value feedback extension as jsonschema doesn't support it
def extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for prop, subschema in properties.items():
            if "default" in subschema and instance is not None:
                instance.setdefault(prop, subschema["default"])

            for error in validate_properties(
                validator, properties, instance, schema
            ):
                yield error

    return validators.extend(validator_class,
                             {"properties": set_defaults})


# Parameter strict check extension as jsonschema doesn't support it
def extend_with_strict_schema(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def strict_schema(validator, properties, instance, schema):
        if instance is None:
            return

        for para in instance.keys():
            if para not in properties.keys():
                raise exception.AuditParameterNotAllowed(parameter=para)

            for error in validate_properties(
                validator, properties, instance, schema
            ):
                yield error

    return validators.extend(validator_class, {"properties": strict_schema})


StrictDefaultValidatingDraft4Validator = extend_with_default(
    extend_with_strict_schema(validators.Draft4Validator))

Draft4Validator = validators.Draft4Validator


def random_string(n):
    return ''.join([random.choice(
        string.ascii_letters + string.digits) for i in range(n)])


# Some clients (e.g. MAAS) use asyncio, which isn't compatible with Eventlet.
# As a workaround, we're delegating such calls to a native thread.
def async_compat_call(f, *args, **kwargs):
    timeout = kwargs.pop('timeout', None)

    async def async_wrapper():
        ret = f(*args, **kwargs)
        if inspect.isawaitable(ret):
            return await asyncio.wait_for(ret, timeout)
        return ret

    def tpool_wrapper():
        # This will run in a separate native thread. Ideally, there should be
        # a single thread permanently running an asyncio loop, but for
        # convenience we'll use eventlet.tpool, which leverages a thread pool.
        #
        # That being considered, we're setting up a temporary asyncio loop to
        # handle this call.
        loop = asyncio.new_event_loop()
        try:
            asyncio.set_event_loop(loop)
            return loop.run_until_complete(async_wrapper())
        finally:
            loop.close()

    # We'll use eventlet timeouts as an extra precaution and asyncio timeouts
    # to avoid lingering threads. For consistency, we'll convert eventlet
    # timeout exceptions to asyncio timeout errors.
    with eventlet.timeout.Timeout(
            seconds=timeout,
            exception=asyncio.TimeoutError("Timeout: %ss" % timeout)):
        return tpool.execute(tpool_wrapper)



# Copyright (c) 2023 Cloudbase Solutions
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock
import uuid

from watcher.common.metal_helper import constants as m_constants


def get_mock_metal_node(node_id=None,
                        power_state=m_constants.PowerState.ON,
                        running_vms=0,
                        hostname=None,
                        compute_state='up'):
    node_id = node_id or str(uuid.uuid4())
    # NOTE(lpetrut): the hostname is important for some of the tests,
    # which expect it to match the fake cluster model.
    hostname = hostname or "compute-" + str(uuid.uuid4()).split('-')[0]

    hypervisor_node_dict = {
        'hypervisor_hostname': hostname,
        'running_vms': running_vms,
        'service': {
            'host': hostname,
        },
        'state': compute_state,
    }
    hypervisor_node = mock.Mock(**hypervisor_node_dict)
    hypervisor_node.to_dict.return_value = hypervisor_node_dict

    node = mock.Mock()
    node.get_power_state.return_value = power_state
    node.get_id.return_value = uuid
    node.get_hypervisor_node.return_value = hypervisor_node
    return node



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# NOTE(sean-k-mooney): watcher does not split up the tests that need eventlet
# and those that do not currently so we need to monkey patch all the tests.
# as an example the watcher.test.cmd module is importing watcher.cmd,
# that has the side effect of monkey patching the test executor
# after many modules are already imported.
from watcher import eventlet
eventlet.patch()

from watcher import objects  # noqa E402

# NOTE(comstud): Make sure we have all of the objects loaded. We do this
# at module import time, because we may be using mock decorators in our
# tests that run at import time.
objects.register_all()



# Copyright 2023 Cloudbase Solutions
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from watcher.common import clients
from watcher.common.metal_helper import factory
from watcher.common.metal_helper import ironic
from watcher.common.metal_helper import maas
from watcher.tests import base


class TestMetalHelperFactory(base.TestCase):

    @mock.patch.object(clients, 'OpenStackClients')
    @mock.patch.object(maas, 'MaasHelper')
    @mock.patch.object(ironic, 'IronicHelper')
    def test_factory(self, mock_ironic, mock_maas, mock_osc):
        self.assertEqual(
            mock_ironic.return_value,
            factory.get_helper())

        self.config(url="fake_maas_url", group="maas_client")
        self.assertEqual(
            mock_maas.return_value,
            factory.get_helper())



# Copyright 2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from oslo_config import cfg

prometheus_client = cfg.OptGroup(name='prometheus_client',
                                 title='Configuration Options for Prometheus',
                                 help="See https://docs.openstack.org/watcher/"
                                      "latest/datasources/prometheus.html for "
                                      "details on how these options are used.")

PROMETHEUS_CLIENT_OPTS = [
    cfg.StrOpt('host',
               help="The hostname or IP address for the prometheus server."),
    cfg.StrOpt('port',
               help="The port number used by the prometheus server."),
    cfg.StrOpt('fqdn_label',
               default="fqdn",
               help="The label that Prometheus uses to store the fqdn of "
                    "exporters. Defaults to 'fqdn'."),
    cfg.StrOpt('instance_uuid_label',
               default="resource",
               help="The label that Prometheus uses to store the uuid of "
                    "OpenStack instances. Defaults to 'resource'."),
    cfg.StrOpt('username',
               help="The basic_auth username to use to authenticate with the "
                    "Prometheus server."),
    cfg.StrOpt('password',
               secret=True,
               help="The basic_auth password to use to authenticate with the "
                    "Prometheus server."),
    cfg.StrOpt('cafile',
               help="Path to the CA certificate for establishing a TLS "
                    "connection with the Prometheus server."),
    cfg.StrOpt('certfile',
               help="Path to the client certificate for establishing a TLS "
                    "connection with the Prometheus server."),
    cfg.StrOpt('keyfile',
               help="Path to the client key for establishing a TLS "
                    "connection with the Prometheus server."),
]


def register_opts(conf):
    conf.register_group(prometheus_client)
    conf.register_opts(PROMETHEUS_CLIENT_OPTS, group=prometheus_client)


def list_opts():
    return [(prometheus_client, PROMETHEUS_CLIENT_OPTS)]



# Copyright 2023 Cloudbase Solutions
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from watcher.common.metal_helper import base
from watcher.common.metal_helper import constants as metal_constants

LOG = log.getLogger(__name__)

POWER_STATES_MAP = {
    'power on': metal_constants.PowerState.ON,
    'power off': metal_constants.PowerState.OFF,
    # For now, we only use ON/OFF states
    'rebooting': metal_constants.PowerState.ON,
    'soft power off': metal_constants.PowerState.OFF,
    'soft reboot': metal_constants.PowerState.ON,
}


class IronicNode(base.BaseMetalNode):
    hv_up_when_powered_off = True

    def __init__(self, ironic_node, nova_node, ironic_client):
        super().__init__(nova_node)

        self._ironic_client = ironic_client
        self._ironic_node = ironic_node

    def get_power_state(self):
        return POWER_STATES_MAP.get(self._ironic_node.power_state,
                                    metal_constants.PowerState.UNKNOWN)

    def get_id(self):
        return self._ironic_node.uuid

    def power_on(self):
        self._ironic_client.node.set_power_state(self.get_id(), "on")

    def power_off(self):
        self._ironic_client.node.set_power_state(self.get_id(), "off")


class IronicHelper(base.BaseMetalHelper):
    @property
    def _client(self):
        if not getattr(self, "_cached_client", None):
            self._cached_client = self._osc.ironic()
        return self._cached_client

    def list_compute_nodes(self):
        out_list = []
        # TODO(lpetrut): consider using "detailed=True" instead of making
        # an additional GET request per node
        node_list = self._client.node.list()

        for node in node_list:
            node_info = self._client.node.get(node.uuid)
            hypervisor_id = node_info.extra.get('compute_node_id', None)
            if hypervisor_id is None:
                LOG.warning('Cannot find compute_node_id in extra '
                            'of ironic node %s', node.uuid)
                continue

            hypervisor_node = self.nova_client.hypervisors.get(hypervisor_id)
            if hypervisor_node is None:
                LOG.warning('Cannot find hypervisor %s', hypervisor_id)
                continue

            out_node = IronicNode(node, hypervisor_node, self._client)
            out_list.append(out_node)

        return out_list

    def get_node(self, node_id):
        ironic_node = self._client.node.get(node_id)
        compute_node_id = ironic_node.extra.get('compute_node_id')
        if compute_node_id:
            compute_node = self.nova_client.hypervisors.get(compute_node_id)
        else:
            compute_node = None
        return IronicNode(ironic_node, compute_node, self._client)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from unittest import mock

import eventlet

from apscheduler.schedulers import background

from watcher.common import scheduling
from watcher import eventlet as eventlet_helper
from watcher.tests import base


class TestSchedulerMonkeyPatching(base.BaseTestCase):

    def setUp(self):
        super().setUp()
        self.started = False
        self.test_scheduler = scheduling.BackgroundSchedulerService()
        self.addCleanup(self._cleanup_scheduler)

    def _cleanup_scheduler(self):
        if self.started:
            self.test_scheduler.shutdown()
            self.started = False

    def _start_scheduler(self):
        self.test_scheduler.start()
        self.started = True

    @mock.patch.object(scheduling.BackgroundSchedulerService, 'start')
    def test_scheduler_start(self, mock_start):
        self.test_scheduler.start()
        mock_start.assert_called_once_with()

    @mock.patch.object(scheduling.BackgroundSchedulerService, 'shutdown')
    def test_scheduler_stop(self, mock_shutdown):
        self._start_scheduler()
        self.test_scheduler.stop()
        mock_shutdown.assert_called_once_with()

    @mock.patch.object(scheduling.BackgroundSchedulerService, '_main_loop')
    def test_scheduler_main_loop(self, mock_main_loop):
        self._start_scheduler()
        mock_main_loop.assert_called_once_with()

    @mock.patch.object(background.BackgroundScheduler, '_main_loop')
    @mock.patch.object(eventlet, 'monkey_patch')
    def test_main_loop_is_monkey_patched(
            self, mock_monky_patch, mock_main_loop):
        self.test_scheduler._main_loop()
        self.assertEqual(
            eventlet_helper.is_patched(), self.test_scheduler.should_patch)
        mock_monky_patch.assert_called_once_with()
        mock_main_loop.assert_called_once_with()

    def test_scheduler_should_patch(self):
        self.assertEqual(
            eventlet_helper.is_patched(), self.test_scheduler.should_patch)



# Copyright (c) 2017 ZTE
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

import jsonschema

from watcher.applier.actions import base as baction
from watcher.applier.actions import change_node_power_state
from watcher.common.metal_helper import constants as m_constants
from watcher.common.metal_helper import factory as m_helper_factory
from watcher.tests import base
from watcher.tests.decision_engine import fake_metal_helper

COMPUTE_NODE = "compute-1"


class TestChangeNodePowerState(base.TestCase):

    def setUp(self):
        super(TestChangeNodePowerState, self).setUp()

        p_m_factory = mock.patch.object(m_helper_factory, 'get_helper')
        m_factory = p_m_factory.start()
        self._metal_helper = m_factory.return_value
        self.addCleanup(p_m_factory.stop)

        # Let's avoid unnecessary sleep calls while running the test.
        p_sleep = mock.patch('time.sleep')
        p_sleep.start()
        self.addCleanup(p_sleep.stop)

        self.input_parameters = {
            baction.BaseAction.RESOURCE_ID: COMPUTE_NODE,
            "state": m_constants.PowerState.ON.value,
        }
        self.action = change_node_power_state.ChangeNodePowerState(
            mock.Mock())
        self.action.input_parameters = self.input_parameters

    def test_parameters_down(self):
        self.action.input_parameters = {
            baction.BaseAction.RESOURCE_ID: COMPUTE_NODE,
            self.action.STATE:
                m_constants.PowerState.OFF.value}
        self.assertTrue(self.action.validate_parameters())

    def test_parameters_up(self):
        self.action.input_parameters = {
            baction.BaseAction.RESOURCE_ID: COMPUTE_NODE,
            self.action.STATE:
                m_constants.PowerState.ON.value}
        self.assertTrue(self.action.validate_parameters())

    def test_parameters_exception_wrong_state(self):
        self.action.input_parameters = {
            baction.BaseAction.RESOURCE_ID: COMPUTE_NODE,
            self.action.STATE: 'error'}
        self.assertRaises(jsonschema.ValidationError,
                          self.action.validate_parameters)

    def test_parameters_resource_id_empty(self):
        self.action.input_parameters = {
            self.action.STATE:
                m_constants.PowerState.ON.value,
        }
        self.assertRaises(jsonschema.ValidationError,
                          self.action.validate_parameters)

    def test_parameters_applies_add_extra(self):
        self.action.input_parameters = {"extra": "failed"}
        self.assertRaises(jsonschema.ValidationError,
                          self.action.validate_parameters)

    def test_change_service_state_pre_condition(self):
        try:
            self.action.pre_condition()
        except Exception as exc:
            self.fail(exc)

    def test_change_node_state_post_condition(self):
        try:
            self.action.post_condition()
        except Exception as exc:
            self.fail(exc)

    def test_execute_node_service_state_with_poweron_target(self):
        self.action.input_parameters["state"] = (
            m_constants.PowerState.ON.value)
        mock_nodes = [
            fake_metal_helper.get_mock_metal_node(
                power_state=m_constants.PowerState.OFF),
            fake_metal_helper.get_mock_metal_node(
                power_state=m_constants.PowerState.ON)
        ]
        self._metal_helper.get_node.side_effect = mock_nodes

        result = self.action.execute()
        self.assertTrue(result)

        mock_nodes[0].set_power_state.assert_called_once_with(
            m_constants.PowerState.ON.value)

    def test_execute_change_node_state_with_poweroff_target(self):
        self.action.input_parameters["state"] = (
            m_constants.PowerState.OFF.value)

        mock_nodes = [
            fake_metal_helper.get_mock_metal_node(
                power_state=m_constants.PowerState.ON),
            fake_metal_helper.get_mock_metal_node(
                power_state=m_constants.PowerState.ON),
            fake_metal_helper.get_mock_metal_node(
                power_state=m_constants.PowerState.OFF)
        ]
        self._metal_helper.get_node.side_effect = mock_nodes

        result = self.action.execute()
        self.assertTrue(result)

        mock_nodes[0].set_power_state.assert_called_once_with(
            m_constants.PowerState.OFF.value)

    def test_revert_change_node_state_with_poweron_target(self):
        self.action.input_parameters["state"] = (
            m_constants.PowerState.ON.value)

        mock_nodes = [
            fake_metal_helper.get_mock_metal_node(
                power_state=m_constants.PowerState.ON),
            fake_metal_helper.get_mock_metal_node(
                power_state=m_constants.PowerState.ON),
            fake_metal_helper.get_mock_metal_node(
                power_state=m_constants.PowerState.OFF)
        ]
        self._metal_helper.get_node.side_effect = mock_nodes

        self.action.revert()

        mock_nodes[0].set_power_state.assert_called_once_with(
            m_constants.PowerState.OFF.value)

    def test_revert_change_node_state_with_poweroff_target(self):
        self.action.input_parameters["state"] = (
            m_constants.PowerState.OFF.value)
        mock_nodes = [
            fake_metal_helper.get_mock_metal_node(
                power_state=m_constants.PowerState.OFF),
            fake_metal_helper.get_mock_metal_node(
                power_state=m_constants.PowerState.ON)
        ]
        self._metal_helper.get_node.side_effect = mock_nodes

        self.action.revert()

        mock_nodes[0].set_power_state.assert_called_once_with(
            m_constants.PowerState.ON.value)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# NOTE(licanwei): Do eventlet monkey patching here, instead of in
# common/service.py.  This allows the API service to run without monkey
# patching under Apache (which uses its own concurrency model). Mixing
# concurrency models can cause undefined behavior and potentially API timeouts.
# NOTE(sean-k-mooney) while ^ is true, since that was written asyncio was added
# to the code base in addition to apscheduler which provides native threads.
# As such we have a lot of technical debt to fix with regards to watchers
# concurrency model as we are mixing up to 3 models the same process.
# apscheduler does not technically support eventlet but it has mostly worked
# until now, apscheduler is used to provide a job schedulers which mixes
# monkey patched and non monkey patched code in the same process.
# That is problematic and can lead to errors on python 3.12+.
# The maas support added asyncio to the codebase which is unsafe to mix
# with eventlets by default.
from watcher import eventlet
eventlet.patch()



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import os


MONKEY_PATCHED = False


def is_patched():
    return MONKEY_PATCHED


def _monkey_patch():
    if is_patched():
        return
    # Anything imported here will not be monkey patched. It is
    # important to take care not to import anything here which requires monkey
    # patching. eventlet processes environment variables at import-time.
    # as such any eventlet configuration should happen here if needed.
    import eventlet
    eventlet.monkey_patch()


def patch():
    # This is only for debugging, this should not be used in production.
    if (os.environ.get('OS_WATCHER_DISABLE_EVENTLET_PATCHING', '').lower()
            not in ('1', 'true', 'yes', 'y')):
        _monkey_patch()
        global MONKEY_PATCHED
        MONKEY_PATCHED = True



# Copyright 2023 Cloudbase Solutions
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from watcher.common import clients
from watcher.common.metal_helper import ironic
from watcher.common.metal_helper import maas

CONF = cfg.CONF


def get_helper(osc=None):
    # TODO(lpetrut): consider caching this client.
    if not osc:
        osc = clients.OpenStackClients()

    if CONF.maas_client.url:
        return maas.MaasHelper(osc)
    else:
        return ironic.IronicHelper(osc)



# Copyright 2023 Cloudbase Solutions
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from watcher.common import exception
from watcher.common.metal_helper import constants as metal_constants


class BaseMetalNode(abc.ABC):
    hv_up_when_powered_off = False

    def __init__(self, nova_node=None):
        self._nova_node = nova_node

    def get_hypervisor_node(self):
        if not self._nova_node:
            raise exception.Invalid(message="No associated hypervisor.")
        return self._nova_node

    def get_hypervisor_hostname(self):
        return self.get_hypervisor_node().hypervisor_hostname

    @abc.abstractmethod
    def get_power_state(self):
        # TODO(lpetrut): document the following methods
        pass

    @abc.abstractmethod
    def get_id(self):
        """Return the node id provided by the bare metal service."""
        pass

    @abc.abstractmethod
    def power_on(self):
        pass

    @abc.abstractmethod
    def power_off(self):
        pass

    def set_power_state(self, state):
        state = metal_constants.PowerState(state)
        if state == metal_constants.PowerState.ON:
            self.power_on()
        elif state == metal_constants.PowerState.OFF:
            self.power_off()
        else:
            raise exception.UnsupportedActionType(
                "Cannot set power state: %s" % state)


class BaseMetalHelper(abc.ABC):
    def __init__(self, osc):
        self._osc = osc

    @property
    def nova_client(self):
        if not getattr(self, "_nova_client", None):
            self._nova_client = self._osc.nova()
        return self._nova_client

    @abc.abstractmethod
    def list_compute_nodes(self):
        pass

    @abc.abstractmethod
    def get_node(self, node_id):
        pass



# Copyright 2023 Cloudbase Solutions
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log

from watcher.common import exception
from watcher.common.metal_helper import base
from watcher.common.metal_helper import constants as metal_constants
from watcher.common import utils

CONF = cfg.CONF
LOG = log.getLogger(__name__)

try:
    from maas.client import enum as maas_enum
except ImportError:
    maas_enum = None


class MaasNode(base.BaseMetalNode):
    hv_up_when_powered_off = False

    def __init__(self, maas_node, nova_node, maas_client):
        super().__init__(nova_node)

        self._maas_client = maas_client
        self._maas_node = maas_node

    def get_power_state(self):
        maas_state = utils.async_compat_call(
            self._maas_node.query_power_state,
            timeout=CONF.maas_client.timeout)

        # python-libmaas may not be available, so we'll avoid a global
        # variable.
        power_states_map = {
            maas_enum.PowerState.ON: metal_constants.PowerState.ON,
            maas_enum.PowerState.OFF: metal_constants.PowerState.OFF,
            maas_enum.PowerState.ERROR: metal_constants.PowerState.ERROR,
            maas_enum.PowerState.UNKNOWN: metal_constants.PowerState.UNKNOWN,
        }
        return power_states_map.get(maas_state,
                                    metal_constants.PowerState.UNKNOWN)

    def get_id(self):
        return self._maas_node.system_id

    def power_on(self):
        LOG.info("Powering on MAAS node: %s %s",
                 self._maas_node.fqdn,
                 self._maas_node.system_id)
        utils.async_compat_call(
            self._maas_node.power_on,
            timeout=CONF.maas_client.timeout)

    def power_off(self):
        LOG.info("Powering off MAAS node: %s %s",
                 self._maas_node.fqdn,
                 self._maas_node.system_id)
        utils.async_compat_call(
            self._maas_node.power_off,
            timeout=CONF.maas_client.timeout)


class MaasHelper(base.BaseMetalHelper):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not maas_enum:
            raise exception.UnsupportedError(
                "MAAS client unavailable. Please install python-libmaas.")

    @property
    def _client(self):
        if not getattr(self, "_cached_client", None):
            self._cached_client = self._osc.maas()
        return self._cached_client

    def list_compute_nodes(self):
        out_list = []
        node_list = utils.async_compat_call(
            self._client.machines.list,
            timeout=CONF.maas_client.timeout)

        compute_nodes = self.nova_client.hypervisors.list()
        compute_node_map = dict()
        for compute_node in compute_nodes:
            compute_node_map[compute_node.hypervisor_hostname] = compute_node

        for node in node_list:
            hypervisor_node = compute_node_map.get(node.fqdn)
            if not hypervisor_node:
                LOG.info('Cannot find hypervisor %s', node.fqdn)
                continue

            out_node = MaasNode(node, hypervisor_node, self._client)
            out_list.append(out_node)

        return out_list

    def _get_compute_node_by_hostname(self, hostname):
        compute_nodes = self.nova_client.hypervisors.search(
            hostname, detailed=True)
        for compute_node in compute_nodes:
            if compute_node.hypervisor_hostname == hostname:
                return compute_node

    def get_node(self, node_id):
        maas_node = utils.async_compat_call(
            self._client.machines.get, node_id,
            timeout=CONF.maas_client.timeout)
        compute_node = self._get_compute_node_by_hostname(maas_node.fqdn)
        return MaasNode(maas_node, compute_node, self._client)



# Copyright 2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
from observabilityclient import prometheus_client
from oslo_config import cfg
from oslo_log import log
import re
from watcher._i18n import _
from watcher.common import exception
from watcher.decision_engine.datasources import base

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class PrometheusHelper(base.DataSourceBase):
    """PrometheusHelper class for retrieving metrics from Prometheus server

    This class implements the DataSourceBase to allow Watcher to query
    Prometheus as a data source for metrics.
    """

    NAME = 'prometheus'
    METRIC_MAP = dict(host_cpu_usage='node_cpu_seconds_total',
                      host_ram_usage='node_memory_MemAvailable_bytes',
                      host_outlet_temp=None,
                      host_inlet_temp=None,
                      host_airflow=None,
                      host_power=None,
                      instance_cpu_usage='ceilometer_cpu',
                      instance_ram_usage='ceilometer_memory_usage',
                      instance_ram_allocated='instance.memory',
                      instance_l3_cache_usage=None,
                      instance_root_disk_size='instance.disk',
                      )
    AGGREGATES_MAP = dict(mean='avg', max='max', min='min', count='avg')

    def __init__(self):
        """Initialise the PrometheusHelper

        The prometheus helper uses the PrometheusAPIClient provided by
        python-observabilityclient.
        The prometheus_fqdn_instance_map maps the fqdn of each node to the
        Prometheus instance label added to all metrics on that node. When
        making queries to Prometheus we use the instance label to specify
        the node for which metrics are to be retrieved.
        host, port and fqdn_label come from watcher_client
        config. The prometheus_fqdn_label allows override of the required label
        in Prometheus scrape configs that specifies each target's fqdn.
        """
        self.prometheus = self._setup_prometheus_client()
        self.prometheus_fqdn_label = (
            CONF.prometheus_client.fqdn_label
        )
        self.prometheus_fqdn_instance_map = (
            self._build_prometheus_fqdn_instance_map()
        )
        self.prometheus_host_instance_map = (
            self._build_prometheus_host_instance_map()
        )

    def _setup_prometheus_client(self):
        """Initialise the prometheus client with config options

        Use the prometheus_client options in watcher.conf to setup
        the PrometheusAPIClient client object and return it.
        :raises watcher.common.exception.MissingParameter if
                prometheus host or port is not set in the watcher.conf
                under the [prometheus_client] section.
        :raises watcher.common.exception.InvalidParameter if
                the prometheus host or port have invalid format.
        """
        def _validate_host_port(host, port):
            if len(host) > 255:
                return (False, "hostname is too long: '%s'" % host)
            if host[-1] == '.':
                host = host[:-1]
            legal_hostname = re.compile(
                "(?!-)[a-z0-9-]{1,63}(?<!-)$", re.IGNORECASE)
            if not all(legal_hostname.match(host_part)
                       for host_part in host.split(".")):
                return (False, "hostname '%s' failed regex match " % host)
            try:
                assert bool(1 <= int(port) <= 65535)
            except (AssertionError, ValueError):
                return (False, "missing or invalid port number '%s' "
                        % port)
            return (True, "all good")

        _host = CONF.prometheus_client.host
        _port = CONF.prometheus_client.port
        if (not _host or not _port):
            raise exception.MissingParameter(
                message=(_(
                    "prometheus host and port must be set in watcher.conf "
                    "under the [prometheus_client] section. Can't initialise "
                    "the datasource without valid host and port."))
            )
        validated, reason = _validate_host_port(_host, _port)
        if (not validated):
            raise exception.InvalidParameter(
                message=(_(
                    "A valid prometheus host and port are required. The #"
                    "values found in watcher.conf are '%(host)s' '%(port)s'. "
                    "This fails validation for the following reason: "
                    "%(reason)s.")
                    % {'host': _host, 'port': _port, 'reason': reason})
            )
        the_client = prometheus_client.PrometheusAPIClient(
            "%s:%s" % (_host, _port))

        # check if tls options or basic_auth options are set and use them
        prometheus_user = CONF.prometheus_client.username
        prometheus_pass = CONF.prometheus_client.password
        prometheus_ca_cert = CONF.prometheus_client.cafile
        prometheus_client_cert = CONF.prometheus_client.certfile
        prometheus_client_key = CONF.prometheus_client.keyfile
        if (prometheus_ca_cert):
            the_client.set_ca_cert(prometheus_ca_cert)
            if (prometheus_client_cert and prometheus_client_key):
                the_client.set_client_cert(
                    prometheus_client_cert, prometheus_client_key)
        if (prometheus_user and prometheus_pass):
            the_client.set_basic_auth(prometheus_user, prometheus_pass)

        return the_client

    def _build_prometheus_fqdn_instance_map(self):
        """Build the fqdn<-->instance_label mapping needed for queries

        Watcher knows nodes by their hostname. In Prometheus however the
        scrape targets (also known as 'instances') are specified by I.P.
        (or hostname) and port number. This function creates a mapping between
        the fully qualified domain name of each node and the corresponding
        instance label used in the scrape config. This relies on a custom
        'fqdn' label added to Prometheus scrape_configs. Operators can use
        a different custom label instead by setting the prometheus_fqdn_label
        config option under the prometheus_client section of watcher config.
        The built prometheus_fqdn_instance_map is used to match watcher
        node.hostname if watcher stores fqdn and otherwise the
        host_instance_map is used instead.
        :return a dict mapping fqdn to instance label. For example:
                {'marios-env-again.controlplane.domain': '10.1.2.3:9100'}
        """
        prometheus_targets = self.prometheus._get(
            "targets?state=active")['data']['activeTargets']
        # >>> prometheus_targets[0]['labels']
        # {'fqdn': 'marios-env-again.controlplane.domain',
        #  'instance': 'localhost:9100', 'job': 'node'}
        fqdn_instance_map = {
            fqdn: instance for (fqdn, instance) in (
                (target['labels'].get(self.prometheus_fqdn_label),
                 target['labels'].get('instance'))
                for target in prometheus_targets
                if target.get('labels', {}).get(self.prometheus_fqdn_label)
            )
        }
        if not fqdn_instance_map:
            LOG.error(
                "Could not create fqdn instance map from Prometheus "
                "targets config. Prometheus returned the following: %s",
                prometheus_targets
            )
            return {}
        return fqdn_instance_map

    def _build_prometheus_host_instance_map(self):
        """Build the hostname<-->instance_label mapping needed for queries

        The prometheus_fqdn_instance_map has the fully qualified domain name
        for hosts. This will create a duplicate map containing only the host
        name part. Depending on the watcher node.hostname either the
        fqdn_instance_map or the host_instance_map will be used to resolve
        the correct prometheus instance label for queries. In the event the
        fqdn_instance_map keys are not valid fqdn (for example it contains
        hostnames, not fqdn) the host_instance_map cannot be created and
        an empty dictionary is returned with a warning logged.
        :return a dict mapping hostname to instance label. For example:
                {'marios-env-again': 'localhost:9100'}
        """
        if not self.prometheus_fqdn_instance_map:
            LOG.error("Cannot build host_instance_map without "
                      "fqdn_instance_map")
            return {}
        host_instance_map = {
            host: instance for (host, instance) in (
                (fqdn.split('.')[0], inst)
                for fqdn, inst in self.prometheus_fqdn_instance_map.items()
                if '.' in fqdn
            )
        }
        if not host_instance_map:
            LOG.warning("Creating empty host instance map. Are the keys "
                        "in prometheus_fqdn_instance_map valid fqdn?")
            return {}
        return host_instance_map

    def _resolve_prometheus_instance_label(self, node_name):
        """Resolve the prometheus instance label to use in queries

        Given the watcher node.hostname, resolve the prometheus instance
        label for use in queries, first trying the fqdn_instance_map and
        then the host_instance_map (watcher.node_name can be fqdn or hostname).
        If the name is not resolved after the first attempt, rebuild the fqdn
        and host instance maps and try again. This allows for new hosts added
        after the initialisation of the fqdn_instance_map.
        :param node_name: the watcher node.hostname
        :return String for the prometheus instance label and None if not found
        """
        def _query_maps(node):
            return self.prometheus_fqdn_instance_map.get(
                node, self.prometheus_host_instance_map.get(node, None))

        instance_label = _query_maps(node_name)
        # refresh the fqdn and host instance maps and retry
        if not instance_label:
            self.prometheus_fqdn_instance_map = (
                self._build_prometheus_fqdn_instance_map()
            )
            self.prometheus_host_instance_map = (
                self._build_prometheus_host_instance_map()
            )
            instance_label = _query_maps(node_name)

        if not instance_label:
            LOG.error("Cannot query prometheus without instance label. "
                      "Could not resolve %s", node_name)
            return None
        return instance_label

    def _resolve_prometheus_aggregate(self, watcher_aggregate, meter):
        """Resolve the prometheus aggregate using self.AGGREGATES_MAP

        This uses the AGGREGATES_MAP to resolve the correct prometheus
        aggregate to use in queries, from the given watcher aggregate
        """
        if watcher_aggregate == 'count':
            LOG.warning('Prometheus data source does not currently support '
                        ' the count aggregate. Proceeding with mean (avg).')
        promql_aggregate = self.AGGREGATES_MAP.get(watcher_aggregate)
        if not promql_aggregate:
            raise exception.InvalidParameter(
                message=(_("Unknown Watcher aggregate %s. This does not "
                           "resolve to any valid prometheus query aggregate.")
                         % watcher_aggregate)
            )
        return promql_aggregate

    def _build_prometheus_query(self, aggregate, meter, instance_label,
                                period, resource=None):
        """Build and return the prometheus query string with the given args

        This function builds and returns the string query that will be sent
        to the Prometheus server /query endpoint. For host cpu usage we use:

        100 - (avg by (instance)(rate(node_cpu_seconds_total{mode='idle',
                                       instance='some_host'}[300s])) * 100)

        so using prometheus rate function over the specified period, we average
        per instance (all cpus) idle time and then 'everything else' is cpu
        usage time.

        For host memory usage we use:

        (node_memory_MemTotal_bytes{instance='the_host'} -
        avg_over_time(
            node_memory_MemAvailable_bytes{instance='the_host'}[300s]))
            / 1024 / 1024

        So we take total and subtract available memory to determine
        how much is in use. We use the prometheus xxx_over_time functions
        avg/max/min depending on the aggregate with the specified time period.

        :param aggregate: one of the values of self.AGGREGATES_MAP
        :param meter: the name of the Prometheus meter to use
        :param instance_label: the Prometheus instance label (scrape target).
        :param period: the period in seconds for which to query
        :param resource: the resource object for which metrics are requested
        :return: a String containing the Prometheus query
        :raises watcher.common.exception.InvalidParameter if params are None
        :raises watcher.common.exception.InvalidParameter if meter is not
                known or currently supported (prometheus meter name).
        """
        query_args = None
        uuid_label_key = CONF.prometheus_client.instance_uuid_label
        if (meter is None or aggregate is None or instance_label is None or
                period is None):
            raise exception.InvalidParameter(
                message=(_(
                    "Cannot build prometheus query without args. "
                    "You provided: meter %(mtr)s, aggregate %(agg)s, "
                    "instance_label %(inst)s, period %(prd)s")
                    % {'mtr': meter, 'agg': aggregate,
                       'inst': instance_label, 'prd': period})
            )

        if meter == 'node_cpu_seconds_total':
            query_args = (
                "100 - (%s by (instance)(rate(%s"
                "{mode='idle',instance='%s'}[%ss])) * 100)" %
                (aggregate, meter, instance_label, period)
            )
        elif meter == 'node_memory_MemAvailable_bytes':
            query_args = (
                "(node_memory_MemTotal_bytes{instance='%s'} "
                "- %s_over_time(%s{instance='%s'}[%ss])) "
                "/ 1024 / 1024" %
                (instance_label, aggregate, meter,
                 instance_label, period)
            )
        elif meter == 'ceilometer_memory_usage':
            query_args = (
                "%s_over_time(%s{%s='%s'}[%ss])" %
                (aggregate, meter, uuid_label_key, instance_label, period)
            )
        elif meter == 'ceilometer_cpu':
            # We are converting the total cumulative cpu time (ns) to cpu usage
            # percentage so we need to divide between the number of vcpus.
            # As this is a percentage metric, we set a max level of 100. It has
            # been observed in very high usage cases, prometheus reporting
            # values higher that 100 what can lead to unexpected behaviors.
            vcpus = resource.vcpus
            if not vcpus:
                LOG.warning(
                    "instance vcpu count not set for instance %s, assuming 1",
                    instance_label
                )
                vcpus = 1
            query_args = (
                "clamp_max((%s by (instance)(rate(%s{%s='%s'}[%ss]))/10e+8) "
                "*(100/%s), 100)" %
                (aggregate, meter, uuid_label_key, instance_label, period,
                 vcpus)
            )
        else:
            raise exception.InvalidParameter(
                message=(_("Cannot process prometheus meter %s") % meter)
            )

        return query_args

    def check_availability(self):
        """check if Prometheus server is available for queries

         Performs HTTP get on the prometheus API /status/runtimeinfo endpoint.
         The prometheus_client will raise a PrometheuAPIClientError if the
         call is unsuccessful, which is caught here and a warning logged.
        """
        try:
            self.prometheus._get("status/runtimeinfo")
        except prometheus_client.PrometheusAPIClientError:
            LOG.warning(
                "check_availability raised PrometheusAPIClientError. "
                "Is Prometheus server down?"
            )
            return 'not available'
        return 'available'

    def list_metrics(self):
        """Fetch all prometheus metrics from api/v1/label/__name__/values

        The prometheus_client will raise a PrometheuAPIClientError if the
        call is unsuccessful, which is caught here and a warning logged.
        """
        try:
            response = self.prometheus._get("label/__name__/values")
        except prometheus_client.PrometheusAPIClientError:
            LOG.warning(
                "list_metrics raised PrometheusAPIClientError. Is Prometheus"
                "server down?"
            )
            return set()
        return set(response['data'])

    def statistic_aggregation(self, resource=None, resource_type=None,
                              meter_name=None, period=300, aggregate='mean',
                              granularity=300):

        meter = self._get_meter(meter_name)
        query_args = ''
        instance_label = ''

        # For instance resource type, the datasource expects the uuid of the
        # instance to be assigned to a label in the prometheus metrics, with a
        # specific key value.
        if resource_type == 'compute_node':
            instance_label = self._resolve_prometheus_instance_label(
                resource.hostname)
        elif resource_type == 'instance':
            instance_label = resource.uuid
            # For ram_allocated and root_disk size metrics there are no valid
            # values in the prometheus backend store. We rely in the values
            # provided in the vms inventory.
            if meter == 'instance.memory':
                return float(resource.memory)
            elif meter == 'instance.disk':
                return float(resource.disk)
        else:
            LOG.warning(
                "Prometheus data source does not currently support "
                "resource_type %s", resource_type
            )
            return None

        promql_aggregate = self._resolve_prometheus_aggregate(aggregate, meter)
        query_args = self._build_prometheus_query(
            promql_aggregate, meter, instance_label, period, resource
        )
        if not query_args:
            LOG.error("Cannot proceed without valid prometheus query")
            return None

        result = self.query_retry(
            self.prometheus.query, query_args,
            ignored_exc=prometheus_client.PrometheusAPIClientError,
        )

        return float(result[0].value) if result else None

    def statistic_series(self, resource=None, resource_type=None,
                         meter_name=None, start_time=None, end_time=None,
                         granularity=300):
        raise NotImplementedError(
            _('Prometheus helper currently does not support statistic_series. '
              'This can be considered for future enhancement.'))

    def _invert_max_min_aggregate(self, agg):
        """Invert max and min for node/host metric queries from node-exporter

            because we query for 'idle'/'unused' cpu and memory.
            For Watcher 'max cpu used' we query for prometheus 'min idle time'.
            For Watcher 'max memory used' we retrieve min 'unused'/'available'
            memory from Prometheus. This internal function is used exclusively
            by get_host_cpu_usage and get_host_ram_usage.
            :param agg: the metric collection aggregate
            :return: a String aggregate

        """
        if agg == 'max':
            return 'min'
        elif agg == 'min':
            return 'max'
        return agg

    def get_host_cpu_usage(self, resource, period=300,
                           aggregate="mean", granularity=None):
        """Query prometheus for node_cpu_seconds_total

        This calculates the host cpu usage and returns it as a percentage
        The calculation is made by using the cpu 'idle' time, per
        instance (so all CPUs are included). For example the query looks like
        (100 - (avg by (instance)(rate(node_cpu_seconds_total
            {mode='idle',instance='localhost:9100'}[300s])) * 100))
        """
        aggregate = self._invert_max_min_aggregate(aggregate)
        cpu_usage = self.statistic_aggregation(
            resource, 'compute_node',
            'host_cpu_usage', period=period,
            granularity=granularity, aggregate=aggregate)
        return float(cpu_usage) if cpu_usage else None

    def get_host_ram_usage(self, resource, period=300,
                           aggregate="mean", granularity=None):
        aggregate = self._invert_max_min_aggregate(aggregate)
        ram_usage = self.statistic_aggregation(
            resource, 'compute_node',
            'host_ram_usage', period=period,
            granularity=granularity, aggregate=aggregate)
        return float(ram_usage) if ram_usage else None

    def get_instance_ram_usage(self, resource, period=300,
                               aggregate="mean", granularity=None):
        ram_usage = self.statistic_aggregation(
            resource, 'instance',
            'instance_ram_usage', period=period,
            granularity=granularity, aggregate=aggregate)
        return ram_usage

    def get_instance_cpu_usage(self, resource, period=300,
                               aggregate="mean", granularity=None):
        cpu_usage = self.statistic_aggregation(
            resource, 'instance',
            'instance_cpu_usage', period=period,
            granularity=granularity, aggregate=aggregate)
        return cpu_usage

    def get_instance_ram_allocated(self, resource, period=300,
                                   aggregate="mean", granularity=None):
        ram_allocated = self.statistic_aggregation(
            resource, 'instance',
            'instance_ram_allocated', period=period,
            granularity=granularity, aggregate=aggregate)
        return ram_allocated

    def get_instance_root_disk_size(self, resource, period=300,
                                    aggregate="mean", granularity=None):
        root_disk_size = self.statistic_aggregation(
            resource, 'instance',
            'instance_root_disk_size', period=period,
            granularity=granularity, aggregate=aggregate)
        return root_disk_size



# Copyright 2023 Cloudbase Solutions
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import asyncio
import time
from unittest import mock

from watcher.common import utils
from watcher.tests import base


class TestCommonUtils(base.TestCase):
    async def test_coro(self, sleep=0, raise_exc=None):
        time.sleep(sleep)
        if raise_exc:
            raise raise_exc
        return mock.sentinel.ret_val

    def test_async_compat(self):
        ret_val = utils.async_compat_call(self.test_coro)
        self.assertEqual(mock.sentinel.ret_val, ret_val)

    def test_async_compat_exc(self):
        self.assertRaises(
            IOError,
            utils.async_compat_call,
            self.test_coro,
            raise_exc=IOError('fake error'))

    def test_async_compat_timeout(self):
        # Timeout not reached.
        ret_val = utils.async_compat_call(self.test_coro, timeout=10)
        self.assertEqual(mock.sentinel.ret_val, ret_val)

        # Timeout reached.
        self.assertRaises(
            asyncio.TimeoutError,
            utils.async_compat_call,
            self.test_coro,
            sleep=0.5, timeout=0.1)



# Copyright 2024 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
from unittest import mock

from observabilityclient import prometheus_client
from oslo_config import cfg

from watcher.common import exception
from watcher.decision_engine.datasources import prometheus as prometheus_helper
from watcher.tests import base


class TestPrometheusHelper(base.BaseTestCase):

    def setUp(self):
        super(TestPrometheusHelper, self).setUp()
        with mock.patch.object(
                prometheus_client.PrometheusAPIClient, '_get',
                return_value={'data': {'activeTargets': [
                    {'labels': {
                        'fqdn': 'marios-env.controlplane.domain',
                        'instance': '10.0.1.2:9100', 'job': 'node',
                    }},
                    {'labels': {
                        'fqdn': 'marios-env-again.controlplane.domain',
                        'instance': 'localhost:9100', 'job': 'node'
                    }}
                ]}}):
            cfg.CONF.prometheus_client.host = "foobarbaz"
            cfg.CONF.prometheus_client.port = "1234"
            self.helper = prometheus_helper.PrometheusHelper()
        stat_agg_patcher = mock.patch.object(
            self.helper, 'statistic_aggregation',
            spec=prometheus_helper.PrometheusHelper.statistic_aggregation)
        self.mock_aggregation = stat_agg_patcher.start()
        self.addCleanup(stat_agg_patcher.stop)
        self.mock_instance = mock.Mock(
            uuid='uuid-0',
            memory=512,
            disk=2,
            vcpus=2)

    def test_unset_missing_prometheus_host(self):
        cfg.CONF.prometheus_client.port = '123'
        cfg.CONF.prometheus_client.host = None
        self.assertRaisesRegex(
            exception.MissingParameter, 'prometheus host and port must be '
                                        'set in watcher.conf',
            prometheus_helper.PrometheusHelper
        )
        cfg.CONF.prometheus_client.host = ''
        self.assertRaisesRegex(
            exception.MissingParameter, 'prometheus host and port must be '
                                        'set in watcher.conf',
            prometheus_helper.PrometheusHelper
        )

    def test_unset_missing_prometheus_port(self):
        cfg.CONF.prometheus_client.host = 'some.host.domain'
        cfg.CONF.prometheus_client.port = None
        self.assertRaisesRegex(
            exception.MissingParameter, 'prometheus host and port must be '
                                        'set in watcher.conf',
            prometheus_helper.PrometheusHelper
        )
        cfg.CONF.prometheus_client.port = ''
        self.assertRaisesRegex(
            exception.MissingParameter, 'prometheus host and port must be '
                                        'set in watcher.conf',
            prometheus_helper.PrometheusHelper
        )

    def test_invalid_prometheus_port(self):
        cfg.CONF.prometheus_client.host = "hostOK"
        cfg.CONF.prometheus_client.port = "123badPort"
        self.assertRaisesRegex(
            exception.InvalidParameter, "missing or invalid port number "
                                        "'123badPort'",
            prometheus_helper.PrometheusHelper
        )
        cfg.CONF.prometheus_client.port = "123456"
        self.assertRaisesRegex(
            exception.InvalidParameter, "missing or invalid port number "
                                        "'123456'",
            prometheus_helper.PrometheusHelper
        )

    def test_invalid_prometheus_host(self):
        cfg.CONF.prometheus_client.port = "123"
        cfg.CONF.prometheus_client.host = "-badhost"
        self.assertRaisesRegex(
            exception.InvalidParameter, "hostname '-badhost' "
                                        "failed regex match",
            prometheus_helper.PrometheusHelper
        )
        too_long_hostname = ("a" * 256)
        cfg.CONF.prometheus_client.host = too_long_hostname
        self.assertRaisesRegex(
            exception.InvalidParameter, ("hostname is too long: " +
                                         "'" + too_long_hostname + "'"),
            prometheus_helper.PrometheusHelper
        )

    @mock.patch.object(prometheus_client.PrometheusAPIClient, 'query')
    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_prometheus_statistic_aggregation(self, mock_prometheus_get,
                                              mock_prometheus_query):
        mock_node = mock.Mock(
            uuid='1234',
            hostname='marios-env.controlplane.domain')

        expected_cpu_usage = 3.2706140350701673

        mock_prom_metric = mock.Mock(
            labels={'instance': '10.0.1.2:9100'},
            timestamp=1731065985.408,
            value=expected_cpu_usage
        )

        mock_prometheus_query.return_value = [mock_prom_metric]
        mock_prometheus_get.return_value = {'data': {'activeTargets': [
            {'labels': {
                'fqdn': 'marios-env.controlplane.domain',
                'instance': '10.0.1.2:9100', 'job': 'node',
            }}]}}
        helper = prometheus_helper.PrometheusHelper()
        result = helper.statistic_aggregation(
            resource=mock_node,
            resource_type='compute_node',
            meter_name='host_cpu_usage',
            period=300,
            aggregate='mean',
            granularity=300,
        )
        self.assertEqual(expected_cpu_usage, result)
        mock_prometheus_query.assert_called_once_with(
            "100 - (avg by (instance)(rate(node_cpu_seconds_total"
            "{mode='idle',instance='10.0.1.2:9100'}[300s])) * 100)")

    @mock.patch.object(prometheus_client.PrometheusAPIClient, 'query')
    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_get_instance_cpu_usage(self, mock_prometheus_get,
                                    mock_prometheus_query):
        mock_instance = self.mock_instance
        expected_cpu_usage = 13.2706140350701673

        mock_prom_metric = mock.Mock(
            labels={'resource': 'uuid-0'},
            timestamp=1731065985.408,
            value=expected_cpu_usage
        )
        mock_prometheus_query.return_value = [mock_prom_metric]
        helper = prometheus_helper.PrometheusHelper()

        cpu_usage = helper.get_instance_cpu_usage(mock_instance)
        self.assertIsInstance(cpu_usage, float)
        self.assertEqual(expected_cpu_usage, cpu_usage)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, 'query')
    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_get_instance_ram_usage(self, mock_prometheus_get,
                                    mock_prometheus_query):

        mock_instance = self.mock_instance
        expected_ram_usage = 49.86

        mock_prom_metric = mock.Mock(
            labels={'resource': 'uuid-0'},
            timestamp=1731065985.408,
            value=expected_ram_usage
        )
        mock_prometheus_query.return_value = [mock_prom_metric]
        helper = prometheus_helper.PrometheusHelper()

        ram_usage = helper.get_instance_ram_usage(
            mock_instance, period=222, aggregate="max",
            granularity=200)
        self.assertIsInstance(ram_usage, float)
        self.assertEqual(expected_ram_usage, ram_usage)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, 'query')
    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_get_instance_ram_allocated(self, mock_prometheus_get,
                                        mock_prometheus_query):

        mock_instance = self.mock_instance
        helper = prometheus_helper.PrometheusHelper()
        ram_allocated = helper.get_instance_ram_allocated(mock_instance,
                                                          period=222,
                                                          aggregate="max")
        self.assertIsInstance(ram_allocated, float)
        self.assertEqual(512, ram_allocated)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, 'query')
    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_get_instance_root_disk_size(self, mock_prometheus_get,
                                         mock_prometheus_query):

        mock_instance = self.mock_instance
        helper = prometheus_helper.PrometheusHelper()
        disk_size = helper.get_instance_root_disk_size(mock_instance,
                                                       period=331,
                                                       aggregate="avg")
        self.assertIsInstance(disk_size, float)
        self.assertEqual(2, disk_size)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, 'query')
    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_prometheus_stt_agg_instance_cpu_usage(self, mock_prometheus_get,
                                                   mock_prometheus_query):
        mock_instance = self.mock_instance
        expected_cpu_usage = 13.2706140350701673

        mock_prom_metric = mock.Mock(
            labels={'resource': 'uuid-0'},
            timestamp=1731065985.408,
            value=expected_cpu_usage
        )
        mock_prometheus_query.return_value = [mock_prom_metric]
        helper = prometheus_helper.PrometheusHelper()
        result_cpu = helper.statistic_aggregation(
            resource=mock_instance,
            resource_type='instance',
            meter_name='instance_cpu_usage',
            period=300,
            granularity=300,
            aggregate='mean',
        )
        self.assertEqual(expected_cpu_usage, result_cpu)
        self.assertIsInstance(result_cpu, float)
        mock_prometheus_query.assert_called_once_with(
            "clamp_max((avg by (instance)(rate("
            "ceilometer_cpu{resource='uuid-0'}[300s]))"
            "/10e+8) *(100/2), 100)"
        )

    @mock.patch.object(prometheus_client.PrometheusAPIClient, 'query')
    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_prometheus_stt_agg_instance_ram_usage(self, mock_prometheus_get,
                                                   mock_prometheus_query):
        mock_instance = self.mock_instance
        expected_ram_usage = 49.86

        mock_prom_metric = mock.Mock(
            labels={'resource': 'uuid-0'},
            timestamp=1731065985.408,
            value=expected_ram_usage
        )
        mock_prometheus_query.return_value = [mock_prom_metric]
        helper = prometheus_helper.PrometheusHelper()
        result_ram_usage = helper.statistic_aggregation(
            resource=mock_instance,
            resource_type='instance',
            meter_name='instance_ram_usage',
            period=300,
            granularity=300,
            aggregate='mean',
        )
        self.assertEqual(expected_ram_usage, result_ram_usage)
        self.assertIsInstance(result_ram_usage, float)
        mock_prometheus_query.assert_called_with(
            "avg_over_time(ceilometer_memory_usage{resource='uuid-0'}[300s])"
        )

    @mock.patch.object(prometheus_client.PrometheusAPIClient, 'query')
    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_prometheus_stt_agg_instance_root_size(self, mock_prometheus_get,
                                                   mock_prometheus_query):
        mock_instance = self.mock_instance

        helper = prometheus_helper.PrometheusHelper()
        result_disk = helper.statistic_aggregation(
            resource=mock_instance,
            resource_type='instance',
            meter_name='instance_root_disk_size',
            period=300,
            granularity=300,
            aggregate='mean',
        )
        self.assertEqual(2, result_disk)
        self.assertIsInstance(result_disk, float)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, 'query')
    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_prometheus_stt_agg_instance_ram_alloc(self, mock_prometheus_get,
                                                   mock_prometheus_query):
        mock_instance = self.mock_instance

        helper = prometheus_helper.PrometheusHelper()
        result_memory = helper.statistic_aggregation(
            resource=mock_instance,
            resource_type='instance',
            meter_name='instance_ram_allocated',
            period=300,
            granularity=300,
            aggregate='mean',
        )
        self.assertEqual(512, result_memory)
        self.assertIsInstance(result_memory, float)

    def test_statistic_aggregation_metric_unavailable(self):
        self.assertRaisesRegex(
            NotImplementedError, 'does not support statistic_series',
            self.helper.statistic_series
        )

    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_prometheus_list_metrics(self, mock_prometheus_get):
        expected_metrics = set(
            ['go_gc_duration_seconds', 'go_gc_duration_seconds_count',
             'go_gc_duration_seconds_sum', 'go_goroutines',]
        )
        mock_prometheus_get.return_value = {
            'status': 'success', 'data': [
                'go_gc_duration_seconds', 'go_gc_duration_seconds_count',
                'go_gc_duration_seconds_sum', 'go_goroutines',
            ]
        }
        result = self.helper.list_metrics()
        self.assertEqual(expected_metrics, result)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_prometheus_list_metrics_error(self, mock_prometheus_get):
        mock_prometheus_get.side_effect = (
            prometheus_client.PrometheusAPIClientError("nope"))
        result = self.helper.list_metrics()
        self.assertEqual(set(), result)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_prometheus_check_availability(self, mock_prometheus_get):
        mock_prometheus_get.return_value = {
            'status': 'success',
            'data': {
                'startTime': '2024-11-05T12:59:56.962333207Z',
                'CWD': '/prometheus', 'reloadConfigSuccess': True,
                'lastConfigTime': '2024-11-05T12:59:56Z',
                'corruptionCount': 0, 'goroutineCount': 30,
                'GOMAXPROCS': 8, 'GOMEMLIMIT': 9223372036854775807,
                'GOGC': '75', 'GODEBUG': '', 'storageRetention': '15d'
            }
        }
        result = self.helper.check_availability()
        self.assertEqual('available', result)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_prometheus_check_availability_error(self, mock_prometheus_get):
        mock_prometheus_get.side_effect = (
            prometheus_client.PrometheusAPIClientError("nope"))
        result = self.helper.check_availability()
        self.assertEqual('not available', result)

    def test_get_host_cpu_usage(self):
        cpu_use = self.helper.get_host_cpu_usage('someNode', 345, 'mean', 300)
        self.assertIsInstance(cpu_use, float)
        self.mock_aggregation.assert_called_once_with(
            'someNode', 'compute_node', 'host_cpu_usage', period=345,
            granularity=300, aggregate='mean')

    def test_get_host_cpu_usage_none(self):
        self.mock_aggregation.return_value = None
        cpu_use = self.helper.get_host_cpu_usage('someNode', 345, 'mean', 300)
        self.assertIsNone(cpu_use)

    def test_get_host_cpu_usage_max(self):
        cpu_use = self.helper.get_host_cpu_usage('theNode', 223, 'max', 100)
        self.assertIsInstance(cpu_use, float)
        self.mock_aggregation.assert_called_once_with(
            'theNode', 'compute_node', 'host_cpu_usage', period=223,
            granularity=100, aggregate='min')

    def test_get_host_cpu_usage_min(self):
        cpu_use = self.helper.get_host_cpu_usage('theNode', 223, 'min', 100)
        self.assertIsInstance(cpu_use, float)
        self.mock_aggregation.assert_called_once_with(
            'theNode', 'compute_node', 'host_cpu_usage', period=223,
            granularity=100, aggregate='max')

    def test_get_host_ram_usage(self):
        ram_use = self.helper.get_host_ram_usage(
            'anotherNode', 456, 'mean', 300)
        self.assertIsInstance(ram_use, float)
        self.mock_aggregation.assert_called_once_with(
            'anotherNode', 'compute_node', 'host_ram_usage', period=456,
            granularity=300, aggregate='mean')

    def test_get_host_ram_usage_none(self):
        self.mock_aggregation.return_value = None
        ram_use = self.helper.get_host_ram_usage('NOPE', 234, 'mean', 567)
        self.assertIsNone(ram_use, float)
        self.mock_aggregation.assert_called()
        self.mock_aggregation.assert_called_once_with(
            'NOPE', 'compute_node', 'host_ram_usage', period=234,
            granularity=567, aggregate='mean')

    def test_get_host_ram_usage_max(self):
        ram_use = self.helper.get_host_ram_usage(
            'aNode', 456, 'max', 300)
        self.assertIsInstance(ram_use, float)
        self.mock_aggregation.assert_called_once_with(
            'aNode', 'compute_node', 'host_ram_usage', period=456,
            granularity=300, aggregate='min')

    def test_get_host_ram_usage_min(self):
        ram_use = self.helper.get_host_ram_usage(
            'aNode', 456, 'min', 300)
        self.assertIsInstance(ram_use, float)
        self.mock_aggregation.assert_called_once_with(
            'aNode', 'compute_node', 'host_ram_usage', period=456,
            granularity=300, aggregate='max')

    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_build_prometheus_fqdn_host_instance_map(
            self, mock_prometheus_get):
        mock_prometheus_get.return_value = {'data': {'activeTargets': [
            {'labels': {
                'fqdn': 'foo.controlplane.domain',
                'instance': '10.1.2.1:9100', 'job': 'node',
            }},
            {'labels': {
                'fqdn': 'bar.controlplane.domain',
                'instance': '10.1.2.2:9100', 'job': 'node',
            }},
            {'labels': {
                'fqdn': 'baz.controlplane.domain',
                'instance': '10.1.2.3:9100', 'job': 'node',
            }},
        ]}}
        expected_fqdn_map = {'foo.controlplane.domain': '10.1.2.1:9100',
                             'bar.controlplane.domain': '10.1.2.2:9100',
                             'baz.controlplane.domain': '10.1.2.3:9100'}
        expected_host_map = {'foo': '10.1.2.1:9100',
                             'bar': '10.1.2.2:9100',
                             'baz': '10.1.2.3:9100'}
        helper = prometheus_helper.PrometheusHelper()
        self.assertEqual(helper.prometheus_fqdn_instance_map,
                         expected_fqdn_map)
        self.assertEqual(helper.prometheus_host_instance_map,
                         expected_host_map)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_missing_prometheus_fqdn_label(self, mock_prometheus_get):
        mock_prometheus_get.return_value = {'data': {'activeTargets': [
            {'labels': {
                'instance': '10.1.2.1:9100', 'job': 'node',
            }},
            {'labels': {
                'instance': '10.1.2.2:9100', 'job': 'node',
            }},
        ]}}
        helper = prometheus_helper.PrometheusHelper()
        self.assertEqual({}, helper.prometheus_fqdn_instance_map)
        self.assertEqual({}, helper.prometheus_host_instance_map)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_using_hostnames_not_fqdn(self, mock_prometheus_get):
        mock_prometheus_get.return_value = {'data': {'activeTargets': [
            {'labels': {
                'fqdn': 'ena',
                'instance': '10.1.2.1:9100', 'job': 'node',
            }},
            {'labels': {
                'fqdn': 'dyo',
                'instance': '10.1.2.2:9100', 'job': 'node',
            }},
        ]}}
        helper = prometheus_helper.PrometheusHelper()
        expected_fqdn_map = {'ena': '10.1.2.1:9100',
                             'dyo': '10.1.2.2:9100'}
        self.assertEqual(
            helper.prometheus_fqdn_instance_map, expected_fqdn_map)
        self.assertEqual({}, helper.prometheus_host_instance_map)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_override_prometheus_fqdn_label(self, mock_prometheus_get):
        mock_prometheus_get.return_value = {'data': {'activeTargets': [
            {'labels': {
                'custom_fqdn_label': 'foo.controlplane.domain',
                'instance': '10.1.2.1:9100', 'job': 'node',
            }},
            {'labels': {
                'custom_fqdn_label': 'bar.controlplane.domain',
                'instance': '10.1.2.2:9100', 'job': 'node',
            }},
        ]}}
        expected_fqdn_map = {'foo.controlplane.domain': '10.1.2.1:9100',
                             'bar.controlplane.domain': '10.1.2.2:9100'}
        expected_host_map = {'foo': '10.1.2.1:9100',
                             'bar': '10.1.2.2:9100'}
        cfg.CONF.prometheus_client.fqdn_label = 'custom_fqdn_label'
        helper = prometheus_helper.PrometheusHelper()
        self.assertEqual(helper.prometheus_fqdn_instance_map,
                         expected_fqdn_map)
        self.assertEqual(helper.prometheus_host_instance_map,
                         expected_host_map)

    def test_resolve_prometheus_instance_label(self):
        expected_instance_label = '10.0.1.2:9100'
        result = self.helper._resolve_prometheus_instance_label(
            'marios-env.controlplane.domain')
        self.assertEqual(result, expected_instance_label)
        result = self.helper._resolve_prometheus_instance_label(
            'marios-env')
        self.assertEqual(result, expected_instance_label)

    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_resolve_prometheus_instance_label_none(self,
                                                    mock_prometheus_get):
        mock_prometheus_get.return_value = {'data': {'activeTargets': []}}
        result = self.helper._resolve_prometheus_instance_label('nope')
        self.assertIsNone(result)
        mock_prometheus_get.assert_called_once_with("targets?state=active")

    def test_build_prometheus_query_node_cpu_avg_agg(self):
        expected_query = (
            "100 - (avg by (instance)(rate(node_cpu_seconds_total"
            "{mode='idle',instance='a_host'}[111s])) * 100)")
        result = self.helper._build_prometheus_query(
            'avg', 'node_cpu_seconds_total', 'a_host', '111')
        self.assertEqual(result, expected_query)

    def test_build_prometheus_query_node_cpu_max_agg(self):
        expected_query = (
            "100 - (max by (instance)(rate(node_cpu_seconds_total"
            "{mode='idle',instance='b_host'}[444s])) * 100)")
        result = self.helper._build_prometheus_query(
            'max', 'node_cpu_seconds_total', 'b_host', '444')
        self.assertEqual(result, expected_query)

    def test_build_prometheus_query_node_memory_avg_agg(self):
        expected_query = (
            "(node_memory_MemTotal_bytes{instance='c_host'} - avg_over_time"
            "(node_memory_MemAvailable_bytes{instance='c_host'}[555s])) "
            "/ 1024 / 1024")
        result = self.helper._build_prometheus_query(
            'avg', 'node_memory_MemAvailable_bytes', 'c_host', '555')
        self.assertEqual(result, expected_query)

    def test_build_prometheus_query_node_memory_min_agg(self):
        expected_query = (
            "(node_memory_MemTotal_bytes{instance='d_host'} - min_over_time"
            "(node_memory_MemAvailable_bytes{instance='d_host'}[222s])) "
            "/ 1024 / 1024")
        result = self.helper._build_prometheus_query(
            'min', 'node_memory_MemAvailable_bytes', 'd_host', '222')
        self.assertEqual(result, expected_query)

    def test_build_prometheus_query_instance_memory_avg_agg(self):
        expected_query = (
            "avg_over_time(ceilometer_memory_usage{resource='uuid-0'}[555s])"
        )
        result = self.helper._build_prometheus_query(
            'avg', 'ceilometer_memory_usage', 'uuid-0', '555')
        self.assertEqual(result, expected_query)

    def test_build_prometheus_query_instance_memory_min_agg(self):
        expected_query = (
            "min_over_time(ceilometer_memory_usage{resource='uuid-0'}[222s])"
        )
        result = self.helper._build_prometheus_query(
            'min', 'ceilometer_memory_usage', 'uuid-0', '222')
        self.assertEqual(result, expected_query)

    def test_build_prometheus_query_instance_cpu_avg_agg(self):
        expected_query = (
            "clamp_max((avg by (instance)(rate("
            "ceilometer_cpu{resource='uuid-0'}[222s]))"
            "/10e+8) *(100/2), 100)"
        )
        result = self.helper._build_prometheus_query(
            'avg', 'ceilometer_cpu', 'uuid-0', '222',
            resource=self.mock_instance)
        self.assertEqual(result, expected_query)

    def test_build_prometheus_query_instance_cpu_max_agg(self):
        expected_query = (
            "clamp_max((max by (instance)(rate("
            "ceilometer_cpu{resource='uuid-0'}[555s]))"
            "/10e+8) *(100/4), 100)"
        )
        mock_instance = mock.Mock(
            uuid='uuid-0',
            memory=512,
            disk=2,
            vcpus=4)
        result = self.helper._build_prometheus_query(
            'max', 'ceilometer_cpu', 'uuid-0', '555', resource=mock_instance)
        self.assertEqual(result, expected_query)

    def test_build_prometheus_query_error(self):
        self.assertRaisesRegex(
            exception.InvalidParameter, 'Cannot process prometheus meter NOPE',
            self.helper._build_prometheus_query,
            'min', 'NOPE', 'the_host', '222'
        )
        self.assertRaisesRegex(
            exception.InvalidParameter, 'instance_label None, period 333',
            self.helper._build_prometheus_query,
            'min', 'node_cpu_seconds_total', None, '333'
        )

    def test_resolve_prometheus_aggregate_vanilla(self):
        result = self.helper._resolve_prometheus_aggregate('mean', 'foo')
        self.assertEqual(result, 'avg')
        result = self.helper._resolve_prometheus_aggregate('count', 'foo')
        self.assertEqual(result, 'avg')
        result = self.helper._resolve_prometheus_aggregate('max', 'foometric')
        self.assertEqual(result, 'max')
        result = self.helper._resolve_prometheus_aggregate('min', 'barmetric')
        self.assertEqual(result, 'min')

    def test_resolve_prometheus_aggregate_unknown(self):
        self.assertRaisesRegex(
            exception.InvalidParameter, 'Unknown Watcher aggregate NOPE.',
            self.helper._resolve_prometheus_aggregate, 'NOPE', 'some_meter')

    @mock.patch.object(prometheus_client.PrometheusAPIClient, '_get')
    def test_prometheus_query_custom_uuid_label(self, mock_prometheus_get):
        cfg.CONF.prometheus_client.instance_uuid_label = 'custom_uuid_label'
        expected_query = (
            "clamp_max((max by (instance)"
            "(rate(ceilometer_cpu{custom_uuid_label='uuid-0'}[555s]))"
            "/10e+8) *(100/4), 100)"
        )
        mock_instance = mock.Mock(
            uuid='uuid-0',
            memory=512,
            disk=2,
            vcpus=4)
        result = self.helper._build_prometheus_query(
            'max', 'ceilometer_cpu', 'uuid-0', '555', resource=mock_instance)
        self.assertEqual(result, expected_query)
